# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
import os
import bisect
import logging
import simplejson
from threading import RLock

IMPALAD_AUDIT_LOG_PREFIX = 'impala_audit_event_log_'
LOG = logging.getLogger( __name__ )
SEEK_CUR = 1

def synchronized(fn):
  """
  Decorator to provide a java style method level "synchronized" facility.
  """
  def new_fn(self, *args, **kwargs):
    self._audit_logs_lock.acquire()
    try:
      result = fn(self, *args, **kwargs)
      return result
    finally:
      self._audit_logs_lock.release()

  return new_fn

def _get_audit_log_timestamp(audit_log_filename):
  """
  Returns the timestamp embedded in the impala audit log filename. The format
  of the filename is 'impala_audit_event_log_<version>-<timestamp>
  """
  index = audit_log_filename.rfind('-')
  return long(audit_log_filename[index+1:])

class AuditLogNotFoundError(Exception):
  """
  Error raised when a log file name is looked into the list of file names in the
  log directory and its not found in the list. Indicates that the file is
  deleted since we read the file list in the directory. 
  """
  def __init__(self, file_name, next_file_name):
    """
    file_name is the file that is not found in the list of audit files we are
    tracking.
    next_file_name is the name of file in the list whose timestamp comes 
    immediately after the timestamp of file_name 
    """
    self._file_name = file_name
    self._next_file_name = next_file_name
    
  def __str__(self):
    return "File %s not found in list of audit files." % (self._file_name, ) 

  def get_next_file_name(self):
    return self._next_file_name

class ImpaladAuditLogs(object):
  """
  This class maintains list of Impalad audit logs in the order they were
  created and provides helper methods to iterate over them. This is used
  to make sure that audit logs are generated for all the audit events, even
  when this process is not running. When this code is executed it picks up
  from the audit log where processing was last done. 
  """
  def __init__(self, log_dir):
    self._log_dir = log_dir
    self._audit_logs = None
    self._audit_logs_timestamp = None
    self._audit_logs_lock = RLock()
    self._get_all_audit_logs()

  @synchronized
  def _get_all_audit_logs(self):
    """
    Read all impala audit files in the impala audit directory and sorts them
    based on their timestamp. The resulting list is stored as a member of this
    class. 
    """
    self._audit_logs = [name for name in os.listdir(self._log_dir) 
      if name.startswith(IMPALAD_AUDIT_LOG_PREFIX)]
    self._audit_logs.sort(key=_get_audit_log_timestamp)
    self._audit_logs_timestamp = [_get_audit_log_timestamp(file_name) for file_name in self._audit_logs]
    LOG.info("%d audit log files exist in audit log directory" % (len(self._audit_logs), ))
    
  @synchronized
  def _get_log_index(self, file_name):
    """
    Returns index of the log specified by 'file_name' in the list of audit logs.
    If the file_name is not present in the list, throws AuditLogNotFoundError.
    """
    num_logs = len(self._audit_logs_timestamp)
    if num_logs == 0:
      raise AuditLogNotFoundError(file_name, None)
      
    timestamp = _get_audit_log_timestamp(file_name)
    idx = bisect.bisect_left(self._audit_logs_timestamp, timestamp)    
    if idx == num_logs:
      raise AuditLogNotFoundError(file_name, None)
    
    # Also, the file should be in our list. If timestamps don't match 
    # the file doesn't exist      
    if self._audit_logs_timestamp[idx] != timestamp:
      raise AuditLogNotFoundError(file_name, self._audit_logs[idx])
    
    return idx
  
  @synchronized
  def get_log(self, file_name):
    """
    Returns a ImpalaAuditLogFile object for the log file specified by file_name.
    If file_name is None, then the oldest timestamp file in the directory 
    is returned. If there is no audit file in the directory, a NoneAuditLogFile 
    object is returned. 
    """
    if file_name is None:
      if self._audit_logs:
        return ImpalaAuditLogFile(self._log_dir, self._audit_logs[0])
      else:
        return NoneAuditLogFile()
    else:
      log_path = os.path.join(self._log_dir, file_name)
      if os.path.exists(log_path):
        return ImpalaAuditLogFile(self._log_dir, file_name)
      else:
        return NonexistentAuditLogFile(file_name)

  @synchronized
  def get_next_log_file_name(self, prev_file_name):
    """
    Checks if a log file following the prev_file_name specified is present in
    the list. If yes, the name of the file is returned, else None is returned.
    """
    ret = None
    if prev_file_name is None:
      if self._audit_logs:
        ret = self._audit_logs[0]
    elif prev_file_name.startswith(IMPALAD_AUDIT_LOG_PREFIX):
      try:
        idx = self._get_log_index(prev_file_name) + 1
        if idx < len(self._audit_logs):
          ret = self._audit_logs[idx]
      except AuditLogNotFoundError, ex:
        ret = ex.get_next_file_name()

    return ret

  @synchronized
  def append_new_log_to_list(self, file_name):
    """Adds a newly created audit log file to the list of audit log files"""
    # First validate if the filename timestamp follows all existing files,
    # otherwise log a warning and ignore the file
    if file_name.startswith(IMPALAD_AUDIT_LOG_PREFIX):
      timestamp = _get_audit_log_timestamp(file_name)
      last_timestamp = -1
      if self._audit_logs:
        last_timestamp = _get_audit_log_timestamp(self._audit_logs[-1])

      if timestamp >= last_timestamp:
        self._audit_logs.append(file_name)
        self._audit_logs_timestamp.append(timestamp)
      else:
        LOG.warning(
          "A new audit log file (%s) has timestamp older than latest log (%s)." %
          (file_name, self._audit_logs[-1], ))
      
  @synchronized
  def remove_log_from_list(self, file_name):
    """
    Updates the listing of the log directory by removing the file that got deleted
    or moved out of the audit log directory.
    """
    if file_name.startswith(IMPALAD_AUDIT_LOG_PREFIX):
      try:
        idx = self._get_log_index(file_name)
        del self._audit_logs[idx]
        del self._audit_logs_timestamp[idx]
      except AuditLogNotFoundError, ex:
        # Not able to find the file in our list of audit logs. Log a message
        LOG.info("Untracked %s audit log file is removed from filesystem." %
                 (file_name, ))
  
class ImpalaAuditLogFile(object):
  """
  This class wraps up an impala audit log and provides method to seek/read
  from this file.
  """
  def __init__(self, log_dir, file_name):
    self._file_name = file_name
    self._fd = open(os.path.join(log_dir, file_name))
  
  def get_file_name(self):
    return self._file_name
  
  def close(self):
    self._fd.close()
    
  def exists(self):
    return True
  
  def seek(self, offset):
    """
    Perform an absolute seek based on the offset provided. os.SEEK_SET is the
    default optional argument to seek function.
    The function is a no-op if the offset is None.
    """    
    if offset is not None:
      self._fd.seek(offset)
      
  def _clear_eof_log_file(self):
    """
    Clears the eof flag by seeking to current location.
    """
    # seek offset of zero from current location to clear the EOF flag 
    self._fd.seek(0, SEEK_CUR)

  def read_audit_log(self):
    """
    Called when there is content to be read from the audit log file. The method
    reads new lines in the file and returns a tuple (filename, offset, json dict).
    The filename is the name of audit file and offset is the offset of the
    start of next line. If the audit can be sent successfully, then this offset
    is recorded in the WAL file. 
    """
    # We will use while loop to read the lines, the better way to do this:
    #   for line in self._fd
    # doesn't fit our case as to make it efficient it has a read-ahead buffer,
    # which means that the "tell" method returns offset that is ahead of the 
    # audit events we are getting. 
    while True:
      line = self._fd.readline()
      if line:
        if line[-1] == "\n":
          try:
            audit_event = simplejson.loads(line)
            yield self._file_name, self._fd.tell(), audit_event
          except simplejson.JSONDecodeError:
            # A \n terminated Audit log line that cannot be parsed is corrupt
            LOG.warn("Skipping corrupt audit log line: %s" % (line,))
        else:
          # We might read half a line as Impala doesn't flush after each write
          # Seek back to the start of the line (can also use distutils.text_file
          # for all this if needed, it exposes unreadline method)
          self._fd.seek(-len(line), SEEK_CUR)
          break
      else:
        break # eof
    
    # At the end clear EOF file so that we can read again
    self._clear_eof_log_file()

class NonexistentAuditLogFile(object):
  """
  This class represents a log file that used to exist but is no longer present.
  We can run into this case when WAL file contains the name of the file, but the
  audit file is deleted. Alternatively, the log directory was changed and wal
  file got copied but the audit logs weren't. 
  """
  def __init__(self, file_name):
    self._file_name = file_name;

  def close(self):
    pass
    
  def seek(self, offset):
    pass
  
  def get_file_name(self):
    return self._file_name
  
  def read_audit_log(self):
    pass
  
  def exists(self):
    return False

class NoneAuditLogFile(object):
  """
  An empty class to act as placeholder for the case where there is no impala
  audit log. This occurs during first start of Impala when there is no audit
  event in the log file.
  """
  def close(self):
    pass
    
  def seek(self, offset):
    pass
  
  def get_file_name(self):
    return None
  
  def exists(self):
    return False
  
  def read_audit_log(self):
    pass